package ssw.mj.test;

import ssw.mj.Token;

/**
 * Test cases for the <code>Scanner</code> class.
 */
public class ScannerTest extends CompilerTestCase {
	public void testOneToken() {
		init(";");
		
		expectToken(Token.semicolon, 1,  1);
		expectToken(Token.eof,       1,  1);
		
		scanAndVerify();
	}

	public void testTwoTokens() {
		init(";;");
		
		expectToken(Token.semicolon, 1,  1);
		expectToken(Token.semicolon, 1,  2);
		expectToken(Token.eof,       1,  2);
		
		scanAndVerify();
	}

	public void testSpace() {
		init(";  ;");
		
		expectToken(Token.semicolon, 1,  1);
		expectToken(Token.semicolon, 1,  4);
		expectToken(Token.eof,       1,  4);
		
		scanAndVerify();
	}

	public void testTabulator() {
		init(";\t\t;");
		
		expectToken(Token.semicolon, 1,  1);
		expectToken(Token.semicolon, 1,  4);
		expectToken(Token.eof,       1,  4);
		
		scanAndVerify();
	}

	public void testNoToken() {
		init("");
		
		expectToken(Token.eof,       1,  0);
		
		scanAndVerify();
	}

	public void testCRLFLineSeparators() {
		init(";" + CR + LF +
		            " ;" + CR + LF +
		            "  ; ");
		
		expectToken(Token.semicolon, 1,  1);
		expectToken(Token.semicolon, 2,  2);
		expectToken(Token.semicolon, 3,  3);
		expectToken(Token.eof,       3,  4);
		
		scanAndVerify();
	}

	public void testLFLineSeparators() {
		init(";" + LF +
		            " ;" + LF +
		            "  ; ");
		
		expectToken(Token.semicolon, 1,  1);
		expectToken(Token.semicolon, 2,  2);
		expectToken(Token.semicolon, 3,  3);
		expectToken(Token.eof,       3,  4);
		
		scanAndVerify();
	}
	
	public void testInvalidChar1() {
		init(" {} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.none,      1,  3);
		expectError(EX >= 2, 1, 3, "INVALID_CHAR", '');
		expectToken(Token.rbrace,    1,  4);
		expectToken(Token.eof,       1,  5);
		
		scanAndVerify();
	}

	public void testInvalidChar2() {
		init(" {\0} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.none,      1,  3);
		expectError(EX >= 2, 1, 3, "INVALID_CHAR", '\0');
		expectToken(Token.rbrace,    1,  4);
		expectToken(Token.eof,       1,  5);
		
		scanAndVerify();
	}

	public void testInvalidChar3() {
		init(" {&} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.none,      1,  3);
		expectError(EX >= 2, 1, 3, "INVALID_CHAR", '&');
		expectToken(Token.rbrace,    1,  4);
		expectToken(Token.eof,       1,  5);
		
		scanAndVerify();
	}

	public void testInvalidChar4() {
		init(" {|} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.none,      1,  3);
		expectError(EX >= 2, 1, 3, "INVALID_CHAR", '|');
		expectToken(Token.rbrace,    1,  4);
		expectToken(Token.eof,       1,  5);
		
		scanAndVerify();
	}

	public void testInvalidChar5() {
		init(" {!} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.none,      1,  3);
		expectError(EX >= 2, 1, 3, "INVALID_CHAR", '!');
		expectToken(Token.rbrace,    1,  4);
		expectToken(Token.eof,       1,  5);
		
		scanAndVerify();
	}

	public void testIdent() {
		init(" {i I i1 i_ i1I_i} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.ident,     1,  3, "i");
		expectToken(Token.ident,     1,  5, "I");
		expectToken(Token.ident,     1,  7, "i1");
		expectToken(Token.ident,     1, 10, "i_");
		expectToken(Token.ident,     1, 13, "i1I_i");
		expectToken(Token.rbrace,    1, 18);
		expectToken(Token.eof,       1, 19);
		
		scanAndVerify();
	}
	
	public void testSingleIdent() {
		init("i");
		
		expectToken(Token.ident,     1,  1, "i");
		expectToken(Token.eof,       1,  1);
		
		scanAndVerify();
	}

	public void testNumber() {
        init(" {123 2147483647} ");

		expectToken(Token.lbrace,    1,  2);
        expectToken(Token.number,    1,  3, 123);
        expectToken(Token.number,    1,  7, 2147483647);
        expectToken(Token.rbrace,    1, 17);
        expectToken(Token.eof,       1, 18);
        
		scanAndVerify();
	}
	
	public void testSingleNumber() {
        init("123");

        expectToken(Token.number,    1,  1, 123);
        expectToken(Token.eof,       1,  3);
        
		scanAndVerify();
	}

	public void testNegativeNumber() {
        init(" {-123} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.minus,     1,  3);
        expectToken(Token.number,    1,  4, 123);
        expectToken(Token.rbrace,    1,  7);
        expectToken(Token.eof,       1,  8);
        
		scanAndVerify();
	}
	
	public void testBigNumber() {
        init(" {2147483648} ");

		expectToken(Token.lbrace,    1,  2);
        expectToken(Token.number,    1,  3, 0);
        expectError(EX >= 2, 1, 3, "BIG_NUM", "2147483648");
        expectToken(Token.rbrace,    1, 13);
        expectToken(Token.eof,       1, 14);
        
		scanAndVerify();
	}
	
	public void testNegatvieBigNumber() {
        init(" {-2147483648} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.minus,     1,  3);
        expectToken(Token.number,    1,  4, 0);
        expectError(EX >= 2, 1, 4, "BIG_NUM", "2147483648");
        expectToken(Token.rbrace,    1, 14);
        expectToken(Token.eof,       1, 15);
        
		scanAndVerify();
	}
	
	public void testReallyBigNumber() {
        init(" {1234567890123456789012345678901234567890} ");

		expectToken(Token.lbrace,    1,  2);
        expectToken(Token.number,    1,  3, 0);
        expectError(EX >= 2, 1, 3, "BIG_NUM", "1234567890123456789012345678901234567890");
        expectToken(Token.rbrace,    1, 43);
        expectToken(Token.eof,       1, 44);
        
		scanAndVerify();
	}
	
	public void testNumberIdent() {
        init(" {123abc123 123while} ");

		expectToken(Token.lbrace,    1,  2);
        expectToken(Token.number,    1,  3, 123);
        expectToken(Token.ident,     1,  6, "abc123");
        expectToken(Token.number,    1, 13, 123);
        expectToken(Token.while_,    1, 16);
        expectToken(Token.rbrace,    1, 21);
        expectToken(Token.eof,       1, 22);
        
		scanAndVerify();
    }

	public void testCharConst() {
		init(" {' ' 'A' 'z' '0' '!' '\"' '' '\0'} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, ' ');
		expectToken(Token.charConst, 1,  7, 'A');
		expectToken(Token.charConst, 1, 11, 'z');
		expectToken(Token.charConst, 1, 15, '0');
		expectToken(Token.charConst, 1, 19, '!');
		expectToken(Token.charConst, 1, 23, '"');
		expectToken(Token.charConst, 1, 27, '');
		expectToken(Token.charConst, 1, 31, '\0');
		expectToken(Token.rbrace,    1, 34);
		expectToken(Token.eof,       1, 35);
		
		scanAndVerify();
	}
	
	public void testSingleCharConst() {
		init("'x'");
		
		expectToken(Token.charConst, 1, 1, 'x');
		expectToken(Token.eof,       1, 3);
		
		scanAndVerify();
	}
	
	public void testEscapeCharConst() {
		init(" {'\\n' '\\r' '\\\\' '\\''} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\n');
		expectToken(Token.charConst, 1,  8, '\r');
		expectToken(Token.charConst, 1, 13, '\\');
		expectToken(Token.charConst, 1, 18, '\'');
		expectToken(Token.rbrace,    1, 22);
		expectToken(Token.eof,       1, 23);
		
		scanAndVerify();
	}
	
	public void testSingleEscapeCharConst() {
		init("'\\n'");
		
		expectToken(Token.charConst, 1,  1, '\n');
		expectToken(Token.eof,       1,  4);
		
		scanAndVerify();
	}
	
	public void testEmptyCharConst() {
		init(" {''} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\0');
		expectError(EX >= 2, 1, 3, "EMPTY_CHARCONST");
		expectToken(Token.rbrace,    1,  5);
		expectToken(Token.eof,       1,  6);
		
		scanAndVerify();
	}

	public void testUnclosedCharConst() {
		init(" {'a} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\0');
		expectError(EX >= 2, 1, 3, "MISSING_QUOTE");
		expectToken(Token.rbrace,    1,  5);
		expectToken(Token.eof,       1,  6);
		
		scanAndVerify();
	}
	
	public void testUnclosedEscapeCharConst() {
		init(" {'\\r} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\0');
		expectError(EX >= 2, 1, 3, "MISSING_QUOTE");
		expectToken(Token.rbrace,    1,  6);
		expectToken(Token.eof,       1,  7);
		
		scanAndVerify();
	}
	
	public void testUnclosedBackslashCharConst() {
		init(" {'\\'} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\0');
		expectError(EX >= 2, 1, 3, "MISSING_QUOTE");
		expectToken(Token.rbrace,    1,  6);
		expectToken(Token.eof,       1,  7);
		
		scanAndVerify();
	}
	
	public void testInvalidEscapeCharConst() {
		init(" {'\\a'} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\0');
		expectError(EX >= 2, 1, 3, "UNDEFINED_ESCAPE", 'a');
		expectToken(Token.rbrace,    1,  7);
		expectToken(Token.eof,       1,  8);
		
		scanAndVerify();
	}
	
	public void testFileEndCharConst() {
		init(" {'");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\0');
		expectError(EX >= 2, 1, 3, "MISSING_QUOTE");
		expectToken(Token.eof,       1,  3);
		
		scanAndVerify();
	}
	
	public void testLineEndCharConst() {
		init(" {'" + LF + 
		     "'a'} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.charConst, 1,  3, '\0');
		expectError(EX >= 2, 1, 3, "ILLEGAL_LINE_END");
		expectToken(Token.charConst, 2,  1, 'a');
		expectToken(Token.rbrace,    2,  4);
		expectToken(Token.eof,       2,  5);
		
		scanAndVerify();
	}

	public void testSimpleSingleLineComment() {
		init(" {/* Simple / single * line comment. */} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.rbrace,    1, 40);
		expectToken(Token.eof,       1, 41);
		
		scanAndVerify();
	}
	
	public void testSimpleMultiLineComment() {
		init(" {" + LF +
		     "  /* Simple " + LF +
		     "     / multi * line " + LF +
		     "     comment. */ " + LF +
		     " } ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.rbrace,    5,  2);
		expectToken(Token.eof,       5,  3);
		
		scanAndVerify();
	}

	public void testNestedSingleLineComment() {
		init(" {/* This / is * a /* nested  /* single line */ comment. */*/} ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.rbrace,    1, 62);
		expectToken(Token.eof,       1, 63);		
		
		scanAndVerify();
	}

	public void testNestedMultiLineComment() {
		init(" {" + LF +
		     "  /* This / is * a " + LF +
		     "   /* nested  " + LF +
		     "    /* multi line */" + LF +
		     "    comment. " + LF +
		     "   */" + LF +
		     "  */ " + LF +
		     " } ");

		expectToken(Token.lbrace,    1,  2);
		expectToken(Token.rbrace,    8,  2);
		expectToken(Token.eof,       8,  3);
		
		scanAndVerify();
	}
	
	public void testUnclosedComment() {
		init(" {/* This / is * a /* nested /* unclosed */ comment. */} ");
		
		expectToken(Token.lbrace,    1,  2);
		expectError(EX >= 2, 1, 3, "EOF_IN_COMMENT");
		expectToken(Token.eof,       1, 57);

		scanAndVerify();
	}

	public void testAllTokens() {
		init("anIdentifier 123 'c'" + LF +
		     "+ - * / % == != < <= > >= && || = += -= *= /= %= ++ -- ; , . ( ) [ ] { }" + LF +
		     "break class else final if new print program read return void while" + LF);
		
		expectToken(Token.ident,     1,  1, "anIdentifier");
		expectToken(Token.number,    1, 14, 123);
		expectToken(Token.charConst, 1, 18, 'c');
		expectToken(Token.plus,      2,  1);
		expectToken(Token.minus,     2,  3);
		expectToken(Token.times,     2,  5);
		expectToken(Token.slash,     2,  7);
		expectToken(Token.rem,       2,  9);
		expectToken(Token.eql,       2, 11);
		expectToken(Token.neq,       2, 14);
		expectToken(Token.lss,       2, 17);
		expectToken(Token.leq,       2, 19);
		expectToken(Token.gtr,       2, 22);
		expectToken(Token.geq,       2, 24);
		expectToken(Token.and,       2, 27);
		expectToken(Token.or,        2, 30);
		expectToken(Token.assign,    2, 33);
		expectToken(Token.plusas,    2, 35);
		expectToken(Token.minusas,   2, 38);
		expectToken(Token.timesas,   2, 41);
		expectToken(Token.slashas,   2, 44);
		expectToken(Token.remas,     2, 47);
		expectToken(Token.pplus,     2, 50);
		expectToken(Token.mminus,    2, 53);
		expectToken(Token.semicolon, 2, 56);
		expectToken(Token.comma,     2, 58);
		expectToken(Token.period,    2, 60);
		expectToken(Token.lpar,      2, 62);
		expectToken(Token.rpar,      2, 64);
		expectToken(Token.lbrack,    2, 66);
		expectToken(Token.rbrack,    2, 68);
		expectToken(Token.lbrace,    2, 70);
		expectToken(Token.rbrace,    2, 72);
		expectToken(Token.break_,    3,  1);
		expectToken(Token.class_,    3,  7);
		expectToken(Token.else_,     3, 13);
		expectToken(Token.final_,    3, 18);
		expectToken(Token.if_,       3, 24);
		expectToken(Token.new_,      3, 27);
		expectToken(Token.print,     3, 31);
		expectToken(Token.program,   3, 37);
		expectToken(Token.read,      3, 45);
		expectToken(Token.return_,   3, 50);
		expectToken(Token.void_,     3, 57);
		expectToken(Token.while_,    3, 62);
		expectToken(Token.eof,       4,  0);

		scanAndVerify();
	}
}
